@@ -9,6 +9,7 @@ import re
 
 from partis.utils import (
   odict,
+  cached_property,
   fmt_base_or_type,
   fmt_iterable_or,
   fmt_attr_doc,
@@ -32,6 +33,7 @@ from partis.schema_meta.base import (
   SchemaDefinitionError,
   SchemaValidationError,
   SchemaHint,
+  Bias,
   Loc,
   assert_valid_name,
   is_bool,
@@ -44,6 +46,7 @@ from partis.schema_meta.base import (
   is_schema,
   is_schema_declared,
   is_schema_struct_valued,
+  is_similar_value_type,
   is_evaluated,
   is_valued,
   is_optional,
@@ -299,6 +302,9 @@ class UnionPrim( SchemaPrim ):
           f"`default_case` must be an integer: {default_case}",
           loc = loc )
 
+      # TODO: check that default_val is 'required', since it would be ambiguous
+      # to specify both default_val and default_case
+
       default_case = int(default_case)
 
       if default_case < 0 or default_case >= len(cases):
@@ -422,13 +428,13 @@ class UnionPrim( SchemaPrim ):
     return cls._p_default_case
 
   #-----------------------------------------------------------------------------
-  @property
+  @cached_property
   def default_val( cls ):
     val = super().default_val
 
-    if not is_required( cls.default_case ) and (
-      is_optional( val )
-      or is_required( val ) ):
+    if (
+      ( is_optional(val) or is_required(val) )
+      and not is_required(cls.default_case) ):
 
       case = cls.cases[ cls.default_case ].schema
       val = case.default_val
@@ -439,12 +445,14 @@ class UnionPrim( SchemaPrim ):
         and case.tag_key in val
         and len(cls._p_schemas) == 1 ):
           # tag is not needed when schema is not ambiguous
+          # NOTE: this is an exception to the rule that the default_val includes the
+          # tag, since there cannot be a union directly in another union.
           val.pop( case.tag_key )
 
     return val
 
   #-----------------------------------------------------------------------------
-  @property
+  @cached_property
   def default_eval( cls ):
     val = super().default_eval
 
@@ -466,7 +474,7 @@ class UnionPrim( SchemaPrim ):
     return val
 
   #-----------------------------------------------------------------------------
-  @property
+  @cached_property
   def init_val( cls ):
     val = super().init_val
 
@@ -562,43 +570,47 @@ class UnionPrim( SchemaPrim ):
   #-----------------------------------------------------------------------------
   def encode( cls,
     val = None,
-    loc = None ):
+    loc = None,
+    no_defaults = None ):
 
     val, loc, validate = cls._encode(
       val = val,
-      loc = loc )
+      loc = loc,
+      no_defaults = no_defaults )
 
     if not validate:
       return val
 
     val = cls.validate(
       val = val,
-      loc = loc )
+      loc = loc )[0]
+
+    schema = None
+    tag_key = None
+    tag_key_optional = False
 
     if is_bool( val ) and cls._p_bool_prim is not None:
-      return cls._p_bool_prim.schema.encode( val )
+      schema = cls._p_bool_prim.schema
 
-    if is_numeric( val ) and cls._p_numeric_prim is not None:
-      return cls._p_numeric_prim.schema.encode( val )
+    elif is_numeric( val ) and cls._p_numeric_prim is not None:
+      schema = cls._p_numeric_prim.schema
 
-    if is_string( val ) and cls._p_string_prim is not None:
-      return cls._p_string_prim.schema.encode( val )
+    elif is_string( val ) and cls._p_string_prim is not None:
+      schema = cls._p_string_prim.schema
 
-    if is_sequence( val ) and cls._p_list_prim is not None:
-      return cls._p_list_prim.schema.encode( val )
+    elif is_sequence( val ) and cls._p_list_prim is not None:
+      schema = cls._p_list_prim.schema
 
-    if is_mapping( val ) and cls._p_dict_prim is not None:
-      return cls._p_dict_prim.schema.encode( val )
+    elif is_mapping( val ) and cls._p_dict_prim is not None:
+      schema = cls._p_dict_prim.schema
 
-    if is_mapping( val ) and len(cls._p_schemas) > 0:
+    elif is_mapping( val ) and len(cls._p_schemas) > 0:
       tag_key = cls._p_schemas[0].tag_key
 
       if len(cls._p_schemas) == 1:
         # only one schema to try
-        val = cls._p_schemas[0].schema.encode( val )
-
-        if is_mapping( val ):
-          val.pop( tag_key )
+        schema = cls._p_schemas[0].schema
+        tag_key_optional = True
 
       else:
         # requires tag to resolve ambiguity
@@ -622,68 +634,117 @@ class UnionPrim( SchemaPrim ):
 
         schema = cls._p_schemas[ cls._p_schema_tags.index( tag ) ].schema
 
-        val = schema.encode( val )
+    else:
+      raise SchemaValidationError(
+        f"`{cls.__name__}` union case not found for value: {fmt_base_or_type(val)}",
+        loc = None,
+        hints = [ SchemaHint(
+          f"case {i}: `{c.schema.__name__}` <{type(c.schema).__name__}>" )
+          for i,c in enumerate(cls.cases) ] )
 
-      return val
+    val = schema.encode(
+      val = val,
+      no_defaults = no_defaults )
 
-    raise SchemaValidationError(
-      f"Union case not found: {type(val).__name__}",
-      loc = None,
-      hints = [ SchemaHint(
-        f"case {i}: {c.schema}" )
-        for i,c in enumerate(cls.cases) ] )
+    # if no_defaults:
+    #   default_val = cls.default_val
+    #
+    #   if not ( is_optional(default_val) or is_required(default_val) ):
+    #
+    #     if is_similar_value_type(val, default_val) and val == default_val:
+    #       # NOTE: is_similar_value_type is needed to prevent removal of values
+    #       # that are "equal", but have a different type.
+    #       # E.G: 0.0 == False, but float != bool.
+    #       val = None
+    #
+    #   if val is None and (
+    #     is_required(default_val)
+    #     or is_optional(default_val)
+    #     or is_required(cls.default_case)
+    #     or schema is not cls.cases[ cls.default_case ].schema ):
+    #
+    #     val = schema.default_val
+
+    if is_mapping(val) and tag_key in val and tag_key_optional:
+      val.pop( tag_key )
+
+    return val
 
   #-----------------------------------------------------------------------------
   def decode( cls,
     val = None,
-    loc = None ):
+    loc = None,
+    bias = None ):
 
-    val, loc, validate = cls._decode(
+    bias = Bias(bias)
+
+    val, loc, validate, bias = cls._decode(
       val = val,
-      loc = loc )
+      loc = loc,
+      bias = bias )
 
     if not validate:
       return val
 
-    val = cls.validate(
+    val, bias = cls.validate(
       val = val,
-      loc = loc )
+      loc = loc,
+      bias = bias )
 
     if is_bool( val ) and cls._p_bool_prim is not None:
       if cls._p_bool_prim in cls._p_proxies:
         return cls._p_proxy_cases[cls._p_proxies.index(cls._p_bool_prim)].schema.decode(
           val = val,
-          loc = loc )
+          loc = loc,
+          bias = bias )
       else:
-        return cls._p_bool_prim.schema.decode( val = val, loc = loc )
+        return cls._p_bool_prim.schema.decode(
+          val = val,
+          loc = loc,
+          bias = bias  )
 
     if is_numeric( val ) and cls._p_numeric_prim is not None:
       if cls._p_numeric_prim in cls._p_proxies:
         return cls._p_proxy_cases[cls._p_proxies.index(cls._p_numeric_prim)].schema.decode(
           val = val,
-          loc = loc )
+          loc = loc,
+          bias = bias  )
       else:
-        return cls._p_numeric_prim.schema.decode( val = val, loc = loc )
+        return cls._p_numeric_prim.schema.decode(
+          val = val,
+          loc = loc,
+          bias = bias )
 
     if is_string( val ) and cls._p_string_prim is not None:
       if cls._p_string_prim in cls._p_proxies:
         return cls._p_proxy_cases[cls._p_proxies.index(cls._p_string_prim)].schema.decode(
           val = val,
-          loc = loc )
+          loc = loc,
+          bias = bias )
       else:
-        return cls._p_string_prim.schema.decode( val = val, loc = loc )
+        return cls._p_string_prim.schema.decode(
+          val = val,
+          loc = loc,
+          bias = bias )
 
     if is_sequence( val ) and cls._p_list_prim is not None:
       if cls._p_list_prim in cls._p_proxies:
         return cls._p_proxy_cases[cls._p_proxies.index(cls._p_list_prim)].schema.decode(
           val = val,
-          loc = loc )
+          loc = loc,
+          bias = bias )
       else:
-        return cls._p_list_prim.schema.decode( val = val, loc = loc )
+        return cls._p_list_prim.schema.decode(
+          val = val,
+          loc = loc,
+          bias = bias )
 
     if is_mapping( val ) and cls._p_dict_prim is not None:
       if cls._p_dict_prim is not None:
-        return cls._p_dict_prim.schema.decode( val = val, loc = loc )
+        return cls._p_dict_prim.schema.decode(
+          val = val,
+          loc = loc,
+          bias = bias )
 
     if is_mapping( val ) and len(cls._p_schemas) > 0:
 
@@ -691,7 +752,8 @@ class UnionPrim( SchemaPrim ):
         # only one schema to try
         return cls._p_schemas[0].schema.decode(
           val = val,
-          loc = loc )
+          loc = loc,
+          bias = bias )
 
       else:
         # requires tag to resolve ambiguity
@@ -720,11 +782,12 @@ class UnionPrim( SchemaPrim ):
 
         return schema.decode(
           val = val,
-          loc = loc )
+          loc = loc,
+          bias = bias )
 
     raise SchemaValidationError(
-      f"Union case not found: {fmt_base_or_type(val)}",
+      f"`{cls.__name__}` union case not found for value: {fmt_base_or_type(val)}",
       loc = loc,
       hints = [ SchemaHint(
-        f"case {i}: {c.schema}" )
+        f"case {i}: `{c.schema.__name__}` <{type(c.schema).__name__}>" )
         for i,c in enumerate(cls.cases) ] )
