@@ -61,7 +61,7 @@ def dump_prep( val, no_defaults ):
     _val = CommentedMap()
 
     for k, v in val.items():
-      _val[k] = dump_prep( v, no_defaults )
+      _val[dump_prep(k, no_defaults)] = dump_prep( v, no_defaults )
 
   elif is_sequence( val ):
     _val = CommentedSeq()
@@ -123,11 +123,11 @@ def loads(
 
   except MarkedYAMLError as e:
     line = getattr(e.problem_mark, 'line', None)
-    line = None if line is None else line + 1 
+    line = None if line is None else line + 1
 
     col = getattr(e.problem_mark, 'column', None)
-    col = None if col is None else col + 1 
-    
+    col = None if col is None else col + 1
+
     data = None
     if hasattr(e.problem_mark, 'get_snippet'):
       data = e.problem_mark.get_snippet()
