@@ -13,6 +13,7 @@ import logging
 log = logging.getLogger(__name__)
 
 from partis.utils import (
+  cached_property,
   adict,
   odict,
   rdict,
@@ -38,6 +39,7 @@ from partis.utils.special import (
   derived )
 
 from .base import (
+  schema_errors,
   SchemaError,
   SchemaNameError,
   SchemaDeclaredError,
@@ -45,6 +47,7 @@ from .base import (
   SchemaValidationError,
   SchemaHint,
   Loc,
+  PresetValue,
   assert_valid_name,
   assert_valid_path,
   is_bool,
@@ -338,12 +341,11 @@ class SchemaDep( SchemaRef ):
 
     namespace = {
       **namespace,
+      '_p_schema_resolved' : False,
       # all dependencies of this schema
       '_p_schema_deps' : schema_deps,
       # schemas referencing this schema as a dependency
-      '_p_schema_refs' : schema_refs,
-      '_p_schema_hash_node' : None,
-      '_p_schema_hash' : None }
+      '_p_schema_refs' : schema_refs }
 
     cls = super().__new__( mcls, name, bases, namespace )
 
@@ -362,7 +364,7 @@ class SchemaDep( SchemaRef ):
   #-----------------------------------------------------------------------------
   @property
   def schema_defined( cls ):
-    return len(cls._p_schema_deps) == 0
+    return cls._p_schema_resolved and len(cls._p_schema_deps) == 0
 
   #-----------------------------------------------------------------------------
   def schema_depends( cls, schema ):
@@ -393,6 +395,9 @@ class SchemaDep( SchemaRef ):
     schema : :class:`SchemaRef <partis.schema_meta.schema.SchemaRef`
     """
 
+    if cls._p_schema_resolved:
+      return True
+
     if schema:
       if schema in cls._p_schema_deps:
         cls._p_schema_deps.remove( schema )
@@ -405,7 +410,15 @@ class SchemaDep( SchemaRef ):
       return False
 
     # fully resolved
-    cls._schema_resolved()
+    cls._p_schema_resolved = True
+
+    with schema_errors(
+      msg = f"Resolution failed: {cls.__module__}.{cls.__name__}",
+      loc = cls.loc if hasattr(cls, 'loc') else None,
+      cls = SchemaDefinitionError ):
+
+      cls._schema_resolved()
+
     refs = cls._p_schema_refs
     cls._p_schema_refs = list()
 
@@ -413,6 +426,7 @@ class SchemaDep( SchemaRef ):
       # signal schemas referencing this schema that it is now defined
       schema.schema_resolved( cls )
 
+
     return True
 
   #-----------------------------------------------------------------------------
@@ -460,7 +474,7 @@ class SchemaDep( SchemaRef ):
     return list()
 
   #-----------------------------------------------------------------------------
-  @property
+  @cached_property
   def schema_hash_node( cls ):
     """urlsafe Base64 encoded hash of this schema, **not** including connections
     (dependencies) to other schemas.
@@ -477,13 +491,8 @@ class SchemaDep( SchemaRef ):
     str
     """
 
-    if not cls._p_schema_hash_node:
-
-      cls._p_schema_hash_node = hash_sha256(
-        '\n'.join( cls._schema_hash_node() ).encode('utf-8') )[0]
-
-    return cls._p_schema_hash_node
-
+    return hash_sha256(
+      '\n'.join( cls._schema_hash_node() ).encode('utf-8') )[0]
 
   #-----------------------------------------------------------------------------
   def _schema_hash_children( cls ):
@@ -597,7 +606,7 @@ class SchemaDep( SchemaRef ):
     return lines
 
   #-----------------------------------------------------------------------------
-  @property
+  @cached_property
   def schema_hash( cls ):
     """urlsafe Base64 encoded hash of this schema, **including** connections
     to other schemas.
@@ -614,12 +623,8 @@ class SchemaDep( SchemaRef ):
 
     """
 
-    if not cls._p_schema_hash:
-
-      cls._p_schema_hash = hash_sha256(
-        '\n'.join( cls._schema_hash() ).encode('utf-8') )[0]
-
-    return cls._p_schema_hash
+    return hash_sha256(
+      '\n'.join( cls._schema_hash() ).encode('utf-8') )[0]
 
 #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 class SchemaDeclared( SchemaDep ):
@@ -799,6 +804,7 @@ class Schema( SchemaDep ):
     If `None`, `default_val` will be used, or a value will be derived
     from the schema to produce an initial value that will pass validation.
     However, `init_val` may specified independently from `default_val`.
+  preset_vals : None | list[ :class: `PresetValue <partis.schema_meta.base.PresetValue>`]
   schema_deps : None | list[ :class:`SchemaRef <partis.schema_meta.schema.SchemaRef>` ]
     Other schemas which this schema depends upon to be defined.
   schema_refs : None | list[ :class:`SchemaRef <partis.schema_meta.schema.SchemaRef>` ]
@@ -830,6 +836,7 @@ class Schema( SchemaDep ):
     default_val = None,
     default_eval = None,
     init_val = None,
+    preset_vals = None,
     schema_deps = None,
     schema_refs = None,
     valued_type = None,
@@ -912,6 +919,20 @@ class Schema( SchemaDep ):
         f"`init_val` may not be a special value: {init_val}",
         loc = loc )
 
+    if preset_vals is None:
+      preset_vals = list()
+
+    if not isinstance(preset_vals, list):
+      raise SchemaDefinitionError(
+        f"`preset_vals` must be a list: {preset_vals}",
+        loc = loc )
+
+    preset_vals = copy(preset_vals)
+
+    for i,v in enumerate(preset_vals):
+      if not isinstance(v, PresetValue):
+        preset_vals[i] = PresetValue(v)
+
     if valued_type is not None:
       if not isinstance( valued_type, type ):
         raise SchemaDefinitionError(
@@ -943,6 +964,7 @@ class Schema( SchemaDep ):
       _p_default_val = default_val,
       _p_default_eval = default_eval,
       _p_init_val = init_val,
+      _p_preset_vals = preset_vals,
       _p_valued_type = valued_type,
       _p_loc = loc )}
 
@@ -1054,7 +1076,7 @@ class Schema( SchemaDep ):
       cls.doc,
       '', '' ]
 
-    if isinstance( cls, Schema ) and depth < max_depth:
+    if isinstance( cls, Schema ) and cls.schema_defined and depth < max_depth:
       lines.extend( cls._get_attr_doc_lines(
         noindex = noindex,
         depth = depth,
@@ -1105,7 +1127,10 @@ class Schema( SchemaDep ):
       'default_eval',
       'init_val']:
 
-      try:
+      with schema_errors(
+        msg = f"{cls.__name__}.{attr} not valid",
+        loc = cls.loc,
+        cls = SchemaDefinitionError ):
 
         v = getattr( cls.schema, attr )
 
@@ -1115,12 +1140,13 @@ class Schema( SchemaDep ):
 
           cls.schema.decode( v, cls.loc )
 
-      except Exception as e:
-        raise SchemaDefinitionError(
-          f"Schema `{attr}` not valid",
-          loc = cls.loc,
-          hints = SchemaHint.cast( e ) ) from e
+    for i, preset in enumerate(cls.preset_vals):
+      with schema_errors(
+        msg = f"{cls.__name__}.preset_vals[{i}] not valid",
+        loc = cls.loc,
+        cls = SchemaDefinitionError ):
 
+        cls.schema.decode( preset.val, cls.loc )
 
   #-----------------------------------------------------------------------------
   @property
@@ -1144,6 +1170,13 @@ class Schema( SchemaDep ):
     """
     return cls._p_init_val
 
+  #-----------------------------------------------------------------------------
+  @property
+  def preset_vals( cls ):
+    """object : preset values that may be used
+    """
+    return cls._p_preset_vals
+
   #-----------------------------------------------------------------------------
   @property
   def evaluated( cls ):
