@@ -14,6 +14,7 @@ import logging
 log = logging.getLogger(__name__)
 
 from partis.utils import (
+  cached_property,
   adict,
   rdict,
   rlist,
@@ -39,6 +40,7 @@ from partis.utils.special import (
   notset )
 
 from partis.schema_meta.base import (
+  schema_errors,
   SchemaError,
   SchemaNameError,
   SchemaDeclaredError,
@@ -239,6 +241,7 @@ class SchemaStruct( Schema ):
     default_val = None,
     default_eval = None,
     init_val = None,
+    preset_vals = None,
     valued_type = None,
     doc = None,
     loc = None ):
@@ -361,6 +364,8 @@ class SchemaStruct( Schema ):
         f"`evaluated` must be subclass of `Evaluated`: {fmt_base_or_type(evaluated)}",
         loc = loc )
 
+    if is_required(struct_proxy):
+      struct_proxy = None
 
     # evaluated_check = evaluated.check if evaluated else lambda x: False
     if struct_proxy:
@@ -376,13 +381,13 @@ class SchemaStruct( Schema ):
           loc = loc )
 
     #...........................................................................
-    for k, v in struct.items():
-      namespace[k] = SchemaProperty(
-        schema = v,
-        name = k )
-
-    namespace[ tag_key ] = ConstProperty(
-      name = tag_key )
+    # for k, v in struct.items():
+    #   namespace[k] = SchemaProperty(
+    #     schema = v,
+    #     name = k )
+    #
+    # namespace[ tag_key ] = ConstProperty(
+    #   name = tag_key )
 
     namespace = { **namespace, **dict(
       _p_tag_key = tag_key,
@@ -400,6 +405,7 @@ class SchemaStruct( Schema ):
       default_val = default_val,
       default_eval = default_eval,
       init_val = init_val,
+      preset_vals = preset_vals,
       schema_deps = schema_deps,
       valued_type = valued_type,
       doc = doc,
@@ -421,6 +427,7 @@ class SchemaStruct( Schema ):
     default_val = None,
     default_eval = None,
     init_val = None,
+    preset_vals = None,
     valued_type = None,
     doc = None,
     loc = None ):
@@ -526,6 +533,8 @@ class SchemaStruct( Schema ):
 
     if cls.struct_proxy:
       for k, v in cls.struct.items():
+        v = v.schema
+
         if k != cls.struct_proxy and is_required( v.default_val ):
           # NOTE: this only validates that no other items in the struct are required
 
@@ -564,7 +573,7 @@ class SchemaStruct( Schema ):
     return cls._p_struct_proxy
 
   #-----------------------------------------------------------------------------
-  @property
+  @cached_property
   def default_val( cls ):
 
     val = cls._p_default_val
@@ -593,7 +602,7 @@ class SchemaStruct( Schema ):
     if is_derived( val ):
       # attempt to derive a default value
 
-      val = dict()
+      val = { cls.tag_key : cls.tag }
 
       for k, _schema in cls.struct.items():
         _schema = _schema.schema
@@ -618,18 +627,19 @@ class SchemaStruct( Schema ):
           val[k] = _val
 
     # ensure default value includes the tag_key: tag
-    if is_mapping(val):
+    if is_mapping(val) and cls.tag_key not in val:
       val[ cls.tag_key ] = cls.tag
 
-    if is_optional(val) or is_required(val):
-      return val
-    else:
-      return cls.encode( val, cls.loc )
-
+    return val
 
+    # TODO: why was this here?
+    # if is_optional(val) or is_required(val):
+    #   return val
+    # else:
+    #   return cls.encode( val, cls.loc )
 
   #-----------------------------------------------------------------------------
-  @property
+  @cached_property
   def default_eval( cls ):
 
     val = cls._p_default_eval
@@ -651,7 +661,7 @@ class SchemaStruct( Schema ):
       return cls.encode( val, cls.loc )
 
   #-----------------------------------------------------------------------------
-  @property
+  @cached_property
   def init_val( cls ):
     val = cls._p_init_val
 
@@ -690,9 +700,7 @@ class SchemaStruct( Schema ):
 
 
     # attempt to derive an initial value
-    val = dict()
-
-    val[ cls.tag_key ] = cls.tag
+    val = { cls.tag_key : cls.tag }
 
     for k, _schema in cls.struct.items():
       _schema = _schema.schema
@@ -704,14 +712,13 @@ class SchemaStruct( Schema ):
 
     return cls.encode( val, cls.loc )
 
-
   #-----------------------------------------------------------------------------
   @property
   def evaluated( cls ):
     return cls._p_evaluated
 
   #-----------------------------------------------------------------------------
-  @property
+  @cached_property
   def hints( cls ):
 
     # build hints
@@ -769,26 +776,117 @@ class SchemaStruct( Schema ):
   #-----------------------------------------------------------------------------
   def encode( cls,
     val = None,
-    loc = None ):
+    loc = None,
+    no_defaults = None ):
 
-    if not isinstance( val, cls.valued_type ):
-      val = cls.decode(
-        val = val,
-        loc = loc )
+    if is_valued_type( val ):
 
-    if val is None:
+      if loc is None:
+        loc = val._loc
+
+      if val._schema.schema is not cls or is_evaluated( val._src ):
+        val = val._encode
+
+    if loc is None:
+      loc = Loc(path = [cls.__name__])
+
+    if is_optional(val):
+      val = cls.default_val
+
+      if is_required( val ):
+        raise SchemaValidationError(
+          f"Value is required",
+          loc = loc,
+          hints = cls.doc )
+
+      elif is_optional( val ):
+        return None
+
+    if cls.evaluated.check( val ):
       return val
 
-    return val._encode
+    if not is_mapping( val ):
+      if cls.struct_proxy:
+        # uses source data for primary struct key value with remaining defaults
+        val = { cls.struct_proxy : val }
+
+      else:
+        raise SchemaValidationError(
+          f"Value must be a mapping: {type(val).__name__}",
+          loc = loc )
+
+    _val = {
+      cls.tag_key : cls.tag }
+
+    for k, _schema in cls.struct.items():
+      _schema = _schema.schema
+
+      v = val.get(k)
+      _loc = loc(key = k)
+
+      with schema_errors(
+        msg = f"Encode failed",
+        loc = _loc,
+        schema = cls,
+        cls = SchemaValidationError ):
+
+        v = _schema.encode(
+          val = v,
+          loc = _loc,
+          no_defaults = no_defaults )
+
+      if v is not None:
+        if no_defaults and (
+          (
+            v == _schema.default_val
+            and is_similar_value_type(v, _schema.default_val) )
+          or (
+            is_mapping( v )
+            and is_schema_struct( _schema )
+            and len(v) == 1
+            and _schema.tag_key in v ) ):
+
+          pass
+
+        else:
+          if is_mapping( v ) and is_schema_struct( _schema ):
+            # don't need to store tag key since schema is not ambiguous
+            v.pop( _schema.tag_key )
+
+          _val[k] = v
+
+    val = _val
+
+    if no_defaults and (
+      len(val) == 2
+      and cls.struct_proxy
+      and cls.struct_proxy in val ):
+
+      val = val[cls.struct_proxy]
+
+    # if no_defaults and (
+    #   # one item when only the tag_key: tag exist
+    #   len(val) == 1
+    #   # check that equivalent to default value
+    #   or (
+    #     val == cls.default_val
+    #     and all(
+    #       is_similar_value_type(a,b)
+    #       for a,b in zip( val.values(), cls.default_val.values() )) )):
+    #
+    #   val = None
+
+    return val
 
   #-----------------------------------------------------------------------------
   def decode( cls,
     val = None,
-    loc = None ):
+    loc = None,
+    bias = None ):
 
     valued_type = cls.valued_type
 
-    return valued_type( val, loc )
+    return valued_type( val, loc, bias )
 
 #+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 class SchemaStructProxy( SchemaProxy ):
@@ -807,6 +905,7 @@ class SchemaStructProxy( SchemaProxy ):
     default_val = None,
     default_eval = None,
     init_val = None,
+    preset_vals = None,
     doc = None,
     loc = None ):
 
@@ -829,6 +928,7 @@ class SchemaStructProxy( SchemaProxy ):
         default_val = default_val,
         default_eval = default_eval,
         init_val = init_val,
+        preset_vals = preset_vals,
         doc = doc,
         loc = loc )
 
@@ -851,6 +951,7 @@ class SchemaStructProxy( SchemaProxy ):
       default_val = kwargs['default_val']
       default_eval = kwargs['default_eval']
       init_val = kwargs['init_val']
+      preset_vals = kwargs['preset_vals']
       doc = kwargs['doc']
       loc = kwargs['loc']
 
@@ -964,8 +1065,9 @@ class SchemaStructProxy( SchemaProxy ):
       if init_val is None:
         init_val = _base_schema._p_init_val
 
-      if doc is None and not namespace.get('__doc__', None):
-        doc = _base_schema._p_doc
+      # TODO: this would currently make the docs inherit unrelated docstrings
+      # if doc is None and not namespace.get('__doc__', None):
+      #   doc = _base_schema._p_doc
 
     schema_namespace = dict()
 
@@ -988,6 +1090,7 @@ class SchemaStructProxy( SchemaProxy ):
       default_val = default_val,
       default_eval = default_eval,
       init_val = init_val,
+      preset_vals = preset_vals,
       doc = doc,
       loc = loc )
 
